# Generated by Django 5.0.9 on 2024-11-12 09:50

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("database", "0073_delete_usersearchmodelconfig"),
    ]

    operations = [
        migrations.AlterField(
            model_name="conversation",
            name="title",
            field=models.CharField(blank=True, default=None, max_length=500, null=True),
        ),
    ]
